(*
 Translation from external to internal syntax.
 Makro expansion takes place in the first rule of rfb.
 
 author: Christian Mueller
*)

local
    open AbstractSyntax
in

fun resolveFreeBinder env e =
    let
	fun rfb (Var(info,s,x))           = 
	    if x = ~1 then Environment.get (env,info,s) else Var (info,s,x)

	  | rfb (App(info,e1,e2))         = App(info, rfb e1, rfb e2)
	  | rfb (Abstr(info,x,e1,e2))     = Abstr(info,x,rfb e1, rfb e2)
	  | rfb (Pi(info,x,e1,e2))        = Pi(info,x,rfb e1,rfb e2) 
	  | rfb (Plus(info,e1,e2))        = Plus(info,rfb e1,rfb e2) 
	  | rfb (Minus(info,e1,e2))       = Minus(info,rfb e1,rfb e2) 
	  | rfb (Mult(info,e1,e2))        = Mult(info,rfb e1,rfb e2) 
	  | rfb (Less(info,e1,e2))        = Less(info,rfb e1,rfb e2) 
	  | rfb (LessEq(info,e1,e2))      = LessEq(info,rfb e1,rfb e2) 
	  | rfb (Greater(info,e1,e2))     = Greater(info,rfb e1,rfb e2) 
	  | rfb (GreaterEq(info,e1,e2))   = GreaterEq(info,rfb e1,rfb e2) 
	  | rfb (IntEqual(info,e1,e2))    = IntEqual(info,rfb e1,rfb e2) 
	  | rfb (IntNotEqual(info,e1,e2)) = IntNotEqual(info,rfb e1,rfb e2) 
	  | rfb (IntNeg (info,e))         = IntNeg (info,rfb e)
	  | rfb (BoolNeg (info,e))        = BoolNeg(info,rfb e) 
	  | rfb (If(info,e1,e2,e3))       = If(info,rfb e1,rfb e2,rfb e3) 
	  | rfb (VariantSort(info,xs))    = 
		VariantSort(info, map (fn (l,e) => (l,rfb e)) xs)

	  | rfb (Variant(info,s,e1,e2))   = Variant(info,s,rfb e1,rfb e2) 
	  | rfb (Case(info,e,xs))         = 
		Case(info,rfb e, map (fn (s1,s2,e) => (s1,s2,rfb e)) xs) 

	  | rfb (RecordSort(info, xs))    = 
		RecordSort(info, map (fn (l,e) => (l,rfb e)) xs) 

	  | rfb (Record(info,xs))         = 
		Record(info, map (fn (l,e) => (l,rfb e)) xs) 

	  | rfb (Proj (info,e,s))         = Proj(info,rfb e,s) 
	  | rfb (RecursiveSort(info,s,e)) = RecursiveSort(info,s,rfb e) 
	  | rfb (Fold(info,e1,e2))        = Fold(info,rfb e1,rfb e2) 
	  | rfb (Unfold(info,e1,e2))      = Unfold(info,rfb e1,rfb e2) 
	  (* construct Variant with type annotation *)
	  | rfb (Ascribe(info,Variant(_,s,e1,_),T)) = 
		Variant (info,s,rfb e1, rfb T)

	  | rfb (Ascribe(info,e,T))       = Ascribe(info,rfb e, rfb T)
	  | rfb (Fix(info,e))             = Fix(info,rfb e)
	  | rfb (Ref(info,e))             = Ref(info, rfb e)
	  | rfb (RefType(info,e))         = RefType(info, rfb e)
	  | rfb (Deref(info,e))           = Deref(info, rfb e)
	  | rfb (Assign(info,e1,e2))      = Assign(info,rfb e1,rfb e2)
	  | rfb (Let(info,s,e1,e2))       = Let(info,s,rfb e1,rfb e2)
	  | rfb nullary_constructor       = nullary_constructor 
    in
	rfb e
    end
					 
end
