(* -*-sml-*- *)

(* --------------------------------------------------------------
 * This file contains some test cases for the interpreter
 * in "F" mode. All the tests listed in "s_test.f" also work.
 *
 * author: Christian Mueller
 * -------------------------------------------------------------*)     

switch tm f

(* 1. standart example for a polymorphic identity function *)

rdef id = \X.\x:X.x

rdef result_11 = id Int 3
rdef result_12 = id (\/X.X->X) id 

rdef double = \X.\f:X->X.\a:X. f (f a)

rdef result_13 = double Int (\x:Int.x+1) 0


(* 2. simple church encodings of booleans, cf. TAPL p.347-348 *)

def CBool = \/ X. X->X->X

rdef tru = \X.\t:X.\f:X.t
rdef fls = \X.\t:X.\f:X.f

rdef conjunction = \b1:CBool.\b2:CBool.b1 CBool b2 fls 

rdef result_21 = conjunction tru tru
rdef result_22 = conjunction tru fls
rdef result_23 = conjunction fls tru 
rdef result_24 = conjunction fls fls


(* 
   3. church encodings for integers, cf. TAPL p.348-349, 
      and polymoprhic Fixpoint operator 
*)

def CNat = \/ X. (X->X)->X->X

rdef c0 = \X.\s:X->X.\z:X.z
rdef succ = \n:CNat.\X.\s:X->X.\z:X. n X s (s z)
rdef plus = \n:CNat.\m:CNat.\X.\s:X->X.\z:X. n X s (m X s z)
rdef cnat2int = \n:CNat. n Int (\x:Int.x+1) 0

rdef omega = \T1.\T2.\f:(T1->T2)->(T1->T2).
               (\x:mu X.X->(T1->T2).\y:T1. 
                    f ((unfold (mu X.X->(T1->T2)) x) x) y)

rdef Fix = \T1.\T2.\f:(T1->T2)->(T1->T2).
		(omega T1 T2 f) (fold (mu X.X->(T1->T2)) (omega T1 T2 f))

rdef int2cnat = Fix Int CNat (\f:Int->CNat.\x:Int. 
                 if x<=0 then c0 else succ( f(x-1) ))

rdef result_3 = cnat2int( plus (int2cnat 12) (int2cnat 5)  )

(* 
   4. ADT for integer pairs (cf. "fw_test.f" for a pure church encoding of 
      polymporphic pairs )
      Some short notes about the ADT encoding:
      - the client is the place where you can work with the ADT constructs
      - you cannot use a value constructed by the ADT outside the client, 
        i.e. bind "pair.constr 12 23" to a toplevel identifier 
*)

rdef pairADT = \Y.\c:\/X.{constr:Int->Int->X,first:X->Int,second:X->Int}->Y. 
                  c {f:Int,s:Int} 
		    { constr = \x:Int.\y:Int.{f=x,s=y},  
		      first = \pair:{f:Int,s:Int}. pair@f, 
		      second = \pair:{f:Int,s:Int}. pair@s }
		  

rdef client = \X.\pair:{constr:Int->Int->X,first:X->Int,second:X->Int}.
                 (\p:X. (pair@first p) + (pair@second p)) (pair@constr 12 23)

rdef result_4 = pairADT Int client







